use std::collections::HashMap;
#[cfg(not(target_arch = "wasm32"))]
use std::fs::File;
use std::hash::{Hash, Hasher};
#[cfg(not(target_arch = "wasm32"))]
use std::io::{self, BufRead, BufReader, Write};

pub struct Dict {
    words: Vec<String>,
    nb_wild_cards: u8,
    wild_card_char: char,
}

/// Simple dictionnary, nothing exiting here
impl Dict {
    /// Load the the dictionnary
    #[cfg(not(target_arch = "wasm32"))]
    pub fn load() -> io::Result<Self> {
        let file = BufReader::new(File::open("dict.dat")?);
        let words: std::io::Result<Vec<_>> = file
            .lines()
            .enumerate()
            .filter_map(|(i, e)| (i != 0).then(|| e)) // Skip the first line (size of the dict)
            .into_iter()
            .collect();
        let words = words?;
        Ok(Self {
            words,
            nb_wild_cards: 1,
            wild_card_char: '?',
        })
    }

    pub fn load_from_str(string: &str, nb_wild_cards: u8, wild_card_char: char) -> Self {
        Self {
            words: string
                .lines()
                .enumerate()
                .filter_map(|(i, e)| (i != 0).then(|| e.into()))
                .into_iter()
                .collect(),
            nb_wild_cards,
            wild_card_char,
        }
    }

    /*
    /// Find a word in the dict
    fn find(&self, word: &str) -> Option<Vec<String>> {
        self.0
            .iter()
            .find(|word_d| word_d.as_str() == word)
            .map(|word| vec![word.clone()])
    }
    */
}

/// The precomputed hash map. The words are regrouped in vector, indexed by there
/// [`FrequencyHash`].
///
/// Hash maps have an average access complexity in O(1)
pub struct AnagramDict(HashMap<FrequencyHash, Vec<String>>);

impl From<&Dict> for AnagramDict {
    fn from(dict: &Dict) -> Self {
        let mut map = Self(HashMap::<FrequencyHash, Vec<String>>::new());

        #[cfg(not(target_arch = "wasm32"))]
        let mut i = 0;
        #[cfg(not(target_arch = "wasm32"))]
        let len = dict.words.len();
        for word in dict.words.iter() {
            let freq = FrequencyHash::compute(word);
            map.add_word_with_wild_card(freq, word, dict.nb_wild_cards, dict.wild_card_char);
            #[cfg(not(target_arch = "wasm32"))]
            {
                i += 1;
                print!("{}%\r", (100 * i) / len);
                io::stdout().flush().unwrap();
            }
        }
        map
    }
}
impl AnagramDict {
    /// Find all anagrams to a word present in this dict
    pub fn find(&self, word: &str) -> Option<Vec<String>> {
        let freq = FrequencyHash::compute(word);
        self.0.get(&freq).cloned()
    }
    fn add_word_with_wild_card(
        &mut self,
        freq: FrequencyHash,
        word: &str,
        nb_wild_card: u8,
        wild_card_symbole: char,
    ) {
        self.0
            .entry(freq.clone())
            .and_modify(|anagrams| anagrams.push(word.to_string()))
            .or_insert(vec![word.to_string()]);
        if nb_wild_card != 0 {
            for char_ in freq.0.clone().keys() {
                let mut freq = freq.clone();
                freq.remove_one_char(char_.clone());
                freq.add_one_char(wild_card_symbole);
                self.add_word_with_wild_card(freq, word, nb_wild_card - 1, wild_card_symbole);
            }
        }
    }
}

/// The FrequencyHash is an hashable structure that represent the occurences
/// of chars in a word. Two words have the same FrequencyHash if and only if
/// they are anagrams of each other.
///
/// We use the [`char`] type for simplicity. The initial problem was in ascii
/// anyway, but keep in mind that [`char`]s are utf-8 scalar values, and in
/// some alphabet those do not match characters.
#[derive(Debug, PartialEq, Eq, Clone)]
struct FrequencyHash(HashMap<char, u8>);

impl FrequencyHash {
    /// Compute the [`FrequencyHash`] of a slice.
    fn compute(string: &str) -> Self {
        let mut map = HashMap::<char, u8>::new();
        for char_ in string.chars() {
            map.entry(char_)
                .and_modify(|counter| *counter += 1)
                .or_insert(1);
        }
        Self(map)
    }

    /// Convert the [`FrequencyHash`] into an equivalent vector of tupple.
    /// The vector is sorted in order to have a unique vector representation
    /// for each [`FrequencyHash`].
    ///
    /// This value used to compute the actual hash of the FrequencyHash.
    fn get_unique_vec(&self) -> Vec<(char, u8)> {
        let mut vec: Vec<(char, u8)> = self.0.clone().into_iter().collect();
        vec.sort();
        vec
    }

    /// Remove one character from the hashmap
    fn remove_one_char(&mut self, char_: char) {
        self.0.entry(char_).and_modify(|counter| *counter -= 1);
        if let Some(0) = self.0.get(&char_) {
            self.0.remove(&char_);
        }
    }

    /// Add one character to the hashmap
    fn add_one_char(&mut self, char_: char) {
        self.0
            .entry(char_)
            .and_modify(|counter| *counter += 1)
            .or_insert(1);
    }
}

impl Hash for FrequencyHash {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.get_unique_vec().hash(state);
    }
}
